from __future__ import annotations

import os
from typing import Dict, Optional

from pacco.classes_file_based import PackageManagerFileBased
from pacco.clients import LocalClient, NexusFileClient, FileBasedClientAbstract


class Remote:
    package_manager: PackageManagerFileBased = None

    def __init__(self, name: str, remote_type: str, client: FileBasedClientAbstract):
        self.name = name
        self.remote_type = remote_type
        self.package_manager = PackageManagerFileBased(client)

    def __str__(self):
        return "[{}, {}]".format(self.name, self.remote_type)

    @staticmethod
    def create(name: str, serialized: Dict[str, str]) -> Remote:
        raise NotImplementedError()

    def serialize(self) -> Dict[str, str]:
        raise NotImplementedError()


class LocalRemote(Remote):
    def __init__(self, name: str, remote_type: str, path: Optional[str] = "", clean: Optional[bool] = False):
        if path:
            self.__path = os.path.abspath(path)
        else:
            self.__path = ""
        client = LocalClient(self.__path, clean)
        super(LocalRemote, self).__init__(name, remote_type, client)

    @staticmethod
    def create(name: str, serialized: Dict[str, str]) -> LocalRemote:
        if 'path' not in serialized:
            serialized['path'] = ""
        return LocalRemote(name, serialized['remote_type'], serialized['path'])

    def serialize(self) -> Dict[str, str]:
        return {'remote_type': 'local', 'path': self.__path}


class NexusSiteRemote(Remote):
    def __init__(self, name: str, remote_type: str, url: str, username: str, password: str,
                 clean: Optional[bool] = False):
        self.__url = url
        self.__username = username
        self.__password = password
        client = NexusFileClient(url, username, password, clean)
        super(NexusSiteRemote, self).__init__(name, remote_type, client)

    @staticmethod
    def create(name: str, serialized: Dict[str, str]) -> NexusSiteRemote:
        return NexusSiteRemote(name, serialized['remote_type'], serialized['url'],
                               serialized['username'], serialized['password'])

    def serialize(self) -> Dict[str, str]:
        return {'remote_type': 'nexus_site', 'url': self.__url,
                'username': self.__username, 'password': self.__password}



from __future__ import annotations

import re
from typing import List, Tuple, Optional, Dict

from pacco.classes_interface import PackageManager, PackageRegistry, PackageBinary
from pacco.clients import FileBasedClientAbstract


class PackageManagerFileBased(PackageManager):
    """
    An implementation of the PackageManager interface

    Examples:
        >>> from pacco.clients import LocalClient, NexusFileClient
        >>> client = LocalClient(clean=True)
        >>> pm = PackageManagerFileBased(client)
        >>> pm.list_package_registries()
        []
        >>> pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        PR[openssl, compiler, os, version]
        >>> pm.add_package_registry('boost', ['os', 'target', 'type'])
        PR[boost, os, target, type]
        >>> pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        Traceback (most recent call last):
            ...
        FileExistsError: The package registry openssl is already found
        >>> pm.list_package_registries()
        ['boost', 'openssl']
        >>> pm.delete_package_registry('openssl')
        >>> pm.list_package_registries()
        ['boost']
        >>> pm.get_package_registry('boost')
        PR[boost, os, target, type]
    """
    def __init__(self, client: FileBasedClientAbstract):
        if not isinstance(client, FileBasedClientAbstract):
            raise TypeError("Must be using FileBasedClient")
        super(PackageManagerFileBased, self).__init__(client)

    def list_package_registries(self) -> List[str]:
        return sorted(self.client.ls())

    def delete_package_registry(self, name: str) -> None:
        self.client.rmdir(name)

    def add_package_registry(self, name: str, settings_key: List[str]) -> PackageRegistryFileBased:
        dirs = self.client.ls()
        if name in dirs:
            raise FileExistsError("The package registry {} is already found".format(name))
        self.client.mkdir(name)
        return PackageRegistryFileBased(name, self.client.dispatch_subdir(name), settings_key)

    def get_package_registry(self, name: str) -> PackageRegistryFileBased:
        dirs = self.client.ls()
        if name not in dirs:
            raise FileNotFoundError("The package registry {} is not found".format(name))
        return PackageRegistryFileBased(name, self.client.dispatch_subdir(name))

    def __repr__(self):
        return "PackageManagerObject"


class PackageRegistryFileBased(PackageRegistry):
    """
    An implementation of the PackageRegistry interface

    Examples:
        >>> from pacco.clients import LocalClient, NexusFileClient
        >>> client = LocalClient(clean=True)
        >>> pm = PackageManagerFileBased(client)
        >>> pr = pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        >>> pr.list_package_binaries()
        []
        >>> pr.add_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        PackageBinaryObject
        >>> pr.add_package_binary({'host_os':'osx', 'compiler':'clang', 'version':'1.0'})
        Traceback (most recent call last):
            ...
        KeyError: "wrong settings key: ['compiler', 'host_os', 'version'] is not ['compiler', 'os', 'version']"
        >>> pr.add_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        Traceback (most recent call last):
            ...
        FileExistsError: such binary already exist
        >>> len(pr.list_package_binaries())
        1
        >>> pr.add_package_binary({'os':'linux', 'compiler':'gcc', 'version':'1.0'})
        PackageBinaryObject
        >>> len(pr.list_package_binaries())
        2
        >>> pr.delete_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        >>> len(pr.list_package_binaries())
        1
        >>> pr.get_package_binary({'os':'linux', 'compiler':'gcc', 'version':'1.0'})
        PackageBinaryObject
    """

    def __init__(self, name: str, client: FileBasedClientAbstract, settings_key: Optional[List[str]] = None):
        if not isinstance(client, FileBasedClientAbstract):
            raise TypeError("Must be using FileBasedClient")
        super(PackageRegistryFileBased, self).__init__(name, client, settings_key)
        from_remote = self.__get_settings_key()
        if settings_key is None and from_remote is None:
            raise FileNotFoundError("you need to declare settings_key if you are adding, if you are getting, this"
                                    "means that the package registry is not properly set, you need to delete and"
                                    "add again")
        elif from_remote is not None:  # ignore the passed settings_key and use the remote one
            self.settings_key = from_remote
        else:
            self.settings_key = settings_key
            self.client.mkdir(self.__generate_settings_key_dir_name(self.settings_key))

    def __repr__(self):
        return "PR[{}, {}]".format(self.name, ', '.join(sorted(self.settings_key)))

    def __get_settings_key(self) -> Optional[List[str]]:
        settings_key = None
        dirs = self.client.ls()
        for dir_name in dirs:
            if '__settings_key' in dir_name:
                settings_key = dir_name.split('==')[1:]
        return settings_key

    @staticmethod
    def __generate_settings_key_dir_name(settings_key: List[str]) -> str:
        settings_key = sorted(settings_key)
        return '=='.join(['__settings_key'] + settings_key)

    @staticmethod
    def __generate_dir_name_from_settings_value(settings_value: Dict[str, str]) -> str:
        sorted_settings_value_pair = sorted(settings_value.items(), key=lambda x: x[0])
        zipped_assignments = ['='.join(pair) for pair in sorted_settings_value_pair]
        return '=='.join(zipped_assignments)

    @staticmethod
    def __generate_settings_value_from_dir_name(dir_name: str) -> Dict[str, str]:
        if not re.match(r"((\w+=\w+)==)*(\w+=\w+)", dir_name):
            raise ValueError("Invalid dir_name syntax {}".format(dir_name))
        return {arg.split('=')[0]: arg.split('=')[1] for arg in dir_name.split('==')}

    def list_package_binaries(self) -> List[Dict[str, str]]:
        dirs = self.client.ls()
        dirs.remove(self.__generate_settings_key_dir_name(self.settings_key))
        return [PackageRegistryFileBased.__generate_settings_value_from_dir_name(name) for name in dirs]

    def add_package_binary(self, settings_value: Dict[str, str]) -> PackageBinaryFileBased:
        if set(settings_value.keys()) != set(self.settings_key):
            raise KeyError("wrong settings key: {} is not {}".format(sorted(settings_value.keys()),
                                                                     sorted(self.settings_key)))
        dir_name = PackageRegistryFileBased.__generate_dir_name_from_settings_value(settings_value)
        if dir_name in self.client.ls():
            raise FileExistsError("such binary already exist")
        self.client.mkdir(dir_name)
        return PackageBinaryFileBased(self.client.dispatch_subdir(dir_name))

    def delete_package_binary(self, settings_value: Dict[str, str]):
        dir_name = PackageRegistryFileBased.__generate_dir_name_from_settings_value(settings_value)
        self.client.rmdir(dir_name)

    def get_package_binary(self, settings_value: Dict[str, str]) -> PackageBinaryFileBased:
        dir_name = PackageRegistryFileBased.__generate_dir_name_from_settings_value(settings_value)
        if set(settings_value.keys()) != set(self.settings_key):
            raise KeyError("wrong settings key: {} is not {}".format(sorted(settings_value.keys()),
                                                                     sorted(self.settings_key)))
        if dir_name not in self.client.ls():
            raise FileNotFoundError("such configuration does not exist")
        return PackageBinaryFileBased(self.client.dispatch_subdir(dir_name))


class PackageBinaryFileBased(PackageBinary):
    """
    An implementation of the PackageBinary interface

    Examples:
        >>> from pacco.clients import LocalClient, NexusFileClient
        >>> client = LocalClient(clean=True)
        >>> pm = PackageManagerFileBased(client)
        >>> pr = pm.add_package_registry('openssl', ['os', 'compiler', 'version'])
        >>> pb = pr.add_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})
        >>> import os, shutil
        >>> os.makedirs('testfolder', exist_ok=True)
        >>> open('testfolder/testfile', 'w').close()
        >>> pb.upload_content('testfolder')
        >>> __ = shutil.move('testfolder/testfile', 'testfolder/testfile2')
        >>> pb_get = pr.get_package_binary({'os':'osx', 'compiler':'clang', 'version':'1.0'})  # use a new reference
        >>> pb_get.download_content('testfolder')
        >>> sorted(os.listdir('testfolder'))
        ['testfile', 'testfile2']
        >>> shutil.rmtree('testfolder')
    """

    def __init__(self, client: FileBasedClientAbstract):
        if not isinstance(client, FileBasedClientAbstract):
            raise TypeError("Must be using FileBasedClient")
        super(PackageBinaryFileBased, self).__init__(client)

    def __repr__(self):
        return "PackageBinaryObject"

    def download_content(self, download_dir_path: str) -> None:
        self.client.download_dir(download_dir_path)

    def upload_content(self, dir_path: str) -> None:
        self.client.upload_dir(dir_path)



from pacco.classes_file_based import PackageManagerFileBased
from pacco.clients import LocalClient
from pacco.cli.output_stream import OutputStream


class PaccoAPIV1:
    def __init__(self):
        self.__client = LocalClient()
        self.__pm = PackageManagerFileBased(self.__client)
        self.out = OutputStream()

    def download(self, registry: str, path: str, *settings: str):
        """
        Download binary, implemented using Pacco's predefined classes and methods.

        Args:
            registry: name of binary
            path: path to download binary
            settings: configuration of binary to be downloaded
        """

        def get_settings_dict(args):
            d = {}
            for s in args:
                if "=" not in s:
                    raise ValueError("settings must be in the form of <key>=<value>")
                key, value = s.split("=")
                d[key] = value
            return d

        try:
            pr = self.__pm.get_package_registry(registry)
            pb = pr.get_package_binary(get_settings_dict(settings))
            pb.download_content(path)
        except KeyError or ValueError as exc:
            self.out.writeln("{}".format(str(exc)), error=True)


PaccoAPI = PaccoAPIV1



import sys


class OutputStream:
    def __init__(self):
        self._stream = sys.stdout
        self._stream_err = sys.stderr

    def writeln(self, data, error=False):
        self.write(data, newline=True, error=error)

    def _write(self, data, newline=False):
        if newline:
            data = "%s\n" % data
        self._stream.write(data)

    def _write_err(self, data, newline=False):
        if newline:
            data = "%s\n" % data
        self._stream_err.write(data)

    def write(self, data, newline=False, error=False):
        if error:
            self._write_err(data, newline)
        else:
            self._write(data, newline)



import argparse
import inspect
from typing import Callable, Dict

from pacco.cli.pacco_api import PaccoAPI


class CommandManager:
    def __init__(self, pacco_api):
        self.__pacco = pacco_api
        self.__out = pacco_api.out

    def run(self, *args):
        """
        Entry point for executing commands, dispatcher to class methods.
        """
        try:
            command = args[0]
            commands = self.__get_commands()
            method = commands[command]
            remaining_args = args[1:]
            method(remaining_args)
        except KeyError as exc:
            if command in ["-h", "--help"]:
                self.__show_help()
                return
            self.__out.writeln("{} is not a pacco command. See 'pacco --help'.".format(str(exc)), error=True)
        except IndexError:  # commands specified
            self.__show_help()

    def __get_commands(self) -> Dict[str, Callable]:
        result = {}
        for method_name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not method_name.startswith('_') and method_name != "run":
                result[method_name] = method
        return result

    def __show_help(self):
        commands = self.__get_commands()
        max_len = max((len(c) for c in commands)) + 1
        fmt = '  %-{}s'.format(max_len)
        for name in commands:
            print(fmt % name, end="")
            docstring_lines = commands[name].__doc__.split('\n')
            data = []
            for line in docstring_lines:
                line = line.strip()
                data.append(line)
            self.__out.writeln(' '.join(data))
        self.__out.writeln("")
        self.__out.writeln("Pacco commands. Type 'pacco <command> -h' for help")

    def download(self, *args: str):
        """
        Download binary by specifying registry, path and settings.
        """
        parser = argparse.ArgumentParser(prog="pacco download")
        parser.add_argument("registry", help="which registry to download")
        parser.add_argument("path", help="path to download registry")
        parser.add_argument("settings", nargs="+", help="settings for the specified registry")
        args = parser.parse_args(*args)
        self.__pacco.download(args.registry, args.path, *args.settings)


def main(args):
    pacco_api = PaccoAPI()
    CommandManager(pacco_api).run(*args)



from setuptools import setup, find_packages

setup(
    name='pacco',
    version='0.2.3',
    packages=find_packages(),
    author="Kevin Winata",
    author_email="kevinwinatamichael@gmail.com",
    description="A simple package manager (used for prebuilt binary), interoperable with Nexus repository manager",
    url="https://kwinata.github.io/pacco/",
    entry_points={
        'console_scripts': [
            'pacco=pacco.cli.entry_point:run'
        ]
    },
    install_requires=[
        'requests',
        'beautifulsoup4',
        'PyYAML',
    ],
    python_requires='>=3.5',
)
